from casioplot import *
import math
import random
WIDTH = 384
HEIGHT = 192
BLACK = (0, 0, 0)

# Vecteurs à 3 dimensions (x,y,z). Ça sert aussi pour les couleurs (r,g,b) !
class Vec3:
    def __init__(self, x, y, z):
        self.x, self.y, self.z = x, y, z
    def __add__(self, u):
        return Vec3(self.x + u.x, self.y + u.y, self.z + u.z)
    def __sub__(self, u):
        return Vec3(self.x - u.x, self.y - u.y, self.z - u.z)
    def neg(self):
        return Vec3(-self.x, -self.y, -self.z)
    # Produit scalaire
    def dot(self, u):
        return self.x * u.x + self.y * u.y + self.z * u.z
    # Produit et division par une constante
    def mulf(self, s):
        return Vec3(self.x * s, self.y * s, self.z * s)
    def divf(self, s):
        return Vec3(self.x / s, self.y / s, self.z / s)
    # Normalisation
    def normalize(self):
        return self.divf(math.sqrt(self.dot(self)))
    # Conversion vers une couleur (avec saturation au blanc)
    def toColor(self):
        f = lambda v: 255 if v >= 1 else int(v * 255)
        return (f(self.x), f(self.y), f(self.z))

def intersect_ray_sphere(ray_origin, ray_dir, sphere_center, sphere_r):
    """Calcul de l'intersection entre un rayon partant de ray_origin et
       voyageant dans la direction ray_dir, et une sphère. Renvoie t1 et t2,
       les deux temps auxquels le rayon intersecte la sphère."""
    ray_origin -= sphere_center

    a = ray_dir.dot(ray_dir)
    b = 2 * ray_origin.dot(ray_dir)
    c = ray_origin.dot(ray_origin) - sphere_r * sphere_r

    delta = b*b - 4*a*c
    if delta < 0:
        return -1, -1
    sqrt_delta = math.sqrt(delta)

    t1, t2 = (-b + sqrt_delta) / (2*a), (-b - sqrt_delta) / (2*a)
    return min(t1, t2), max(t1, t2)

def sphere_normal(sphere_center, position):
    """Vecteur normal à une sphère à la position donnée."""
    return (position - sphere_center).normalize()

ORIGIN = Vec3(0, 0, 0)

# "Field Of View": angle de vue vertical en degrés et en radians
FOV = 60
FOV_RADIANS = FOV * math.pi / 180
# Distance dans l'espace en 3D entre la caméra (à ORIGIN) et le plan qui
# représente l'écran. La distance est choisie pour que l'écran fasse une
# hauteur de 1 unité dans l'espace en 3D.
SCREEN_DISTANCE = 1 / (2 * math.tan(FOV_RADIANS / 2))

### Objet #1: L'étoile de la mort (Death Star, abbrévié "DS") ###

# Centre et rayon de la sphère
DS_CENTER = Vec3(0, 6, -0.2)
DS_RADIUS = 2.4

# Valeurs aléatoire pour le motif à la surface de la sphère
DS_ALEA = [random.randint(0, 255) for _ in range(25)]

# Couleurs
DS_LIGHT     = Vec3(0.7, 0.65, 0.5)
DS_DARK      = Vec3(0.45, 0.4, 0.3)
DS_SEMILIGHT = (DS_LIGHT + DS_DARK).divf(2)
DS_HIGHLIGHT = (DS_DARK + (DS_LIGHT - DS_DARK).mulf(3))

# Fonction donnant la couleur à la surface de la sphère à la position 3D donnée
def texture_death_star(pos):
    # Conversion des coordonnées cartésiennes x/y/z en sphériques. Comme on ne
    # regarde qu'un côté de la sphère connaître x/z suffit.
    pos = (pos - DS_CENTER).divf(DS_RADIUS)
    phi = math.asin(pos.z)
    theta = math.asin(pos.x / math.cos(phi))
    # Conversion de l'invervalles [-π;π] vers [0:256) pour u et [0:128) pour v
    # (symétrique par rapport au milieu de la hauteur)
    u = (theta + math.pi / 2) / math.pi * 256
    v = abs(phi / math.pi * 256)

    # Pixels aléatoires sur la partie sombre mais réguliers en hauteur
    surf_light_N = 0
    if v <= 0.5:
        surf_light_N = 4
    elif 4 < v <= 60:
        surf_light_N = 4 if (int(v) & 7) == 0 else 64
    if surf_light_N and random.randint(0, surf_light_N) == 0:
        return DS_HIGHLIGHT if u < 144 else DS_SEMILIGHT

    # Bordure au milieu
    if v <= 0.5:
        return DS_DARK.mulf(0.7)
    # Bandes de motifs carrés en surface
    if 4 <= v <= 20:
        return DS_LIGHT if int(u) in DS_ALEA else DS_DARK
    if 23 <= v <= 38:
        return DS_LIGHT if (int(u) + 43) & 255 in DS_ALEA else DS_DARK
    if 41 <= v <= 60:
        return DS_LIGHT if (int(u) + 90) & 255 in DS_ALEA else DS_DARK

    return DS_LIGHT

### Object #2: Laser incrusté dans la sphère ###

# Centre et rayon de la sphère qui est extrudée de l'étoile de la mort
LASER_CENTER = DS_CENTER + Vec3(0.28, -1, 0.28).normalize().mulf(3.48)
LASER_RADIUS = 1.2

# Même affaire que pour texture_death_star()
def texture_laser(pos):
    pos = (pos - LASER_CENTER).divf(LASER_RADIUS)
    phi = math.asin(pos.z)
    theta = math.asin(pos.x / math.cos(phi))
    # Conversion des invervalles [-π;π] vers [-1:1] en se restreignant à la
    # partie visible
    u = (theta + math.pi / 2) / math.pi * 8.533 - 3.533
    v = (phi   + math.pi / 2) / math.pi * 8.533 - 3.533

    LIGHT     = Vec3(0.7, 0.7, 0.50)
    DARK      = Vec3(0.3, 0.3, 0.24)
    SEMILIGHT = (DARK + (LIGHT - DARK).mulf(2))

    r = math.sqrt(u*u + v*v)
    if 0.80 <= r <= 0.85:
        return DARK
    if r <= 0.25:
        return SEMILIGHT
    return LIGHT

### Source de lumière #1 (lumière ambiente) ###

# Intensité de l'illumination
AMBIENT_LIGHT = 0.05

### Souce de lumière #2 (lumière directionnelle): le Soleil ###

# Direction des rayons et intensité de l'illumination
SUN_DIRECTION = Vec3(-1, -0.22, -0.1).normalize()
SUN_INTENSITY = 4.5

### Source de lumière #3 (lumière ponctuelle) ###

# "Front light": Position de la source et intensité de l'illumination
FLIGHT_POSITION = Vec3(-1.5, 2.5, 0)
FLIGHT_INTENSITY = 0.15

### Dessin ###

# Fond noir
for y in range(HEIGHT):
    for x in range(WIDTH):
        set_pixel(x, y, BLACK)
show_screen()

# Étoiles aléatoires
for step in range(10):
    for _ in range(400):
        intensity = random.randint(0, 95)
        color = (intensity, intensity, intensity)
        x = random.randint(0, WIDTH-1)
        y = random.randint(0, HEIGHT-1)
        set_pixel(x, y, color)
    show_screen()

# Lancer de rayon pour tous les pixels
for y in range(HEIGHT):
    for x in range(WIDTH):
        # Direction du rayon partant de ORIGIN et intersectant l'écran au
        # niveau du pixel (x, y)
        ray_dir = Vec3((x - WIDTH / 2) / HEIGHT,
                       SCREEN_DISTANCE,
                       (HEIGHT / 2 - y * 1.05) / HEIGHT).normalize()

        # Si le rayon n'intersecte pas l'étoile de la mort, on passe
        t1_ds, t2_ds = intersect_ray_sphere(ORIGIN, ray_dir, DS_CENTER, DS_RADIUS)
        if t1_ds < 0:
            continue

        # Point de collision, normale à la sphère en ce point, et texture
        collision = ORIGIN + ray_dir.mulf(t1_ds)
        normal = sphere_normal(DS_CENTER, collision)
        texture = texture_death_star

        # Si on touche le laser, on utilise les infos de collision du laser
        t1_laser, t2_laser = intersect_ray_sphere(ORIGIN, ray_dir, LASER_CENTER, LASER_RADIUS)
        if 0 <= t1_laser <= t1_ds <= t2_laser:
            collision = ORIGIN + ray_dir.mulf(t2_laser)
            normal = sphere_normal(LASER_CENTER + (LASER_CENTER - DS_CENTER).mulf(0.3), collision).neg()
            texture = texture_laser

        # Calcul de l'illumination du point de collision:

        # Lumière ambiante
        light_level = AMBIENT_LIGHT
        # Lumière provenant du Soleil
        light_level += max(-normal.dot(SUN_DIRECTION) * SUN_INTENSITY, 0)
        # Lumière provenant de la source ponctuelle sur la gauche
        FLIGHT_DIRECTION = (collision - FLIGHT_POSITION).normalize()
        light_level += max(-normal.dot(FLIGHT_DIRECTION) * FLIGHT_INTENSITY, 0)

        # Couleur finale du pixel: couleur donnée par la texture, multipliée
        # par le niveau d'illumination
        color = texture(collision).mulf(light_level)
        set_pixel(x, y, color.toColor())
    show_screen()
